Package: GOSemSim
Type: Package
Title: GO-terms Semantic Similarity Measures
Version: 2.39.2
Authors@R: c( person(given = "Guangchuang", family = "Yu",             email = "guangchuangyu@gmail.com", role  = c("aut", "cre")),
              person(given = "Alexey",      family = "Stukalov",       email = "astukalov@gmail.com",     role  = "ctb"),
              person(given= "Pingfan",      family = "Guo",            email = "1178431277@qq.com",       role  = "ctb"),
              person(given = "Chuanle",     family = "Xiao",           email = "xiaochuanle@126.com",     role  = "ctb"),
              person(given = "Lluís",       family = "Revilla Sancho", email = "lluis.revilla@gmail.com", role  = "ctb")
             )
Maintainer: Guangchuang Yu <guangchuangyu@gmail.com>
Description: Semantic similarity analysis of Gene Ontology (GO)
        annotations provides a quantitative framework for comparing GO
        terms, gene products, and gene clusters. GOSemSim implements
        widely used information content- and graph-based similarity
        measures, including the methods of Resnik, Schlicker, Jiang,
        Lin, Wang and TCSS. It also provides utilities for preparing
        annotation data and combining term-level similarities into
        gene- and cluster-level scores.
Depends: R (>= 4.2.0)
LinkingTo: Rcpp
Imports: AnnotationDbi, DBI, digest, GO.db, methods, rlang, stats,
        utils, yulab.utils (>= 0.2.3)
Suggests: AnnotationHub, BiocParallel, BiocManager, clusterProfiler,
        DOSE, knitr, org.Hs.eg.db, quarto, readr, testthat, tidyr,
        tidyselect, ROCR
VignetteBuilder: quarto
ByteCompile: true
License: Artistic-2.0
Encoding: UTF-8
URL: https://yulab-smu.top/biomedical-knowledge-mining-book/
BugReports: https://github.com/YuLab-SMU/GOSemSim/issues
Packaged: 2026-06-29 14:19:00 UTC; root
biocViews: Annotation, GO, Clustering, Pathways, Network, Software
Roxygen: list(markdown = TRUE)
Config/roxygen2/version: 8.0.0
Config/pak/sysreqs: cmake make libpng-dev libuv1-dev libssl-dev
        zlib1g-dev
Repository: https://bioc.r-universe.dev
Date/Publication: 2026-06-29 09:06:57 UTC
RemoteUrl: https://github.com/bioc/GOSemSim
RemoteRef: HEAD
RemoteSha: e6caf6daa62ff2442ced58e127a2a0b3d84ea33e
NeedsCompilation: yes
Author: Guangchuang Yu [aut, cre],
  Alexey Stukalov [ctb],
  Pingfan Guo [ctb],
  Chuanle Xiao [ctb],
  Lluís Revilla Sancho [ctb]
Built: R 4.6.1; aarch64-apple-darwin23; 2026-06-29 14:21:54 UTC; unix

# Generated by roxygen2: do not edit by hand

export(buildGOmap)
export(clusterSim)
export(combineScores)
export(geneSim)
export(get_organism)
export(goSim)
export(godata)
export(load_OrgDb)
export(mclusterSim)
export(mgeneSim)
export(mgoSim)
export(parse_gff)
export(read.blast2go)
export(read.gaf)
export(set_auto_update)
export(tcss_cutoff)
export(termSim)
export(wangMethod_internal)
exportClasses(GOSemSimDATA)
importClassesFrom(AnnotationDbi,AnnotationDb)
importFrom(AnnotationDbi,columns)
importFrom(AnnotationDbi,loadDb)
importFrom(AnnotationDbi,toTable)
importFrom(DBI,dbReadTable)
importFrom(GO.db,GO.db)
importFrom(GO.db,GOBPANCESTOR)
importFrom(GO.db,GOBPOFFSPRING)
importFrom(GO.db,GOBPPARENTS)
importFrom(GO.db,GOCCANCESTOR)
importFrom(GO.db,GOCCOFFSPRING)
importFrom(GO.db,GOCCPARENTS)
importFrom(GO.db,GOMFANCESTOR)
importFrom(GO.db,GOMFOFFSPRING)
importFrom(GO.db,GOMFPARENTS)
importFrom(GO.db,GOTERM)
importFrom(digest,digest)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setMethod)
importFrom(methods,setRefClass)
importFrom(methods,slot)
importFrom(rlang,check_installed)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(utils,read.delim)
importFrom(utils,setTxtProgressBar)
importFrom(utils,stack)
importFrom(utils,txtProgressBar)
importFrom(yulab.utils,download_yulab_file)
importFrom(yulab.utils,load_OrgDb)
importFrom(yulab.utils,ls2df)
importFrom(yulab.utils,user_dir)
importFrom(yulab.utils,yulab_msg)
importMethodsFrom(AnnotationDbi,as.list)
importMethodsFrom(AnnotationDbi,columns)
importMethodsFrom(AnnotationDbi,dbconn)
importMethodsFrom(AnnotationDbi,keys)
importMethodsFrom(AnnotationDbi,keytypes)
importMethodsFrom(AnnotationDbi,mget)
importMethodsFrom(AnnotationDbi,select)
importMethodsFrom(AnnotationDbi,toTable)
useDynLib(GOSemSim)

#' Measure similarities between two term vectors.
#'
#' Provide two term vectors, this function calculates their similarities.
#'
#' @title termSim
#' @param t1 Term vector
#' @param t2 Term vector
#' @param semData GOSemSimDATA object
#' @param method One of "Wang", "Resnik", "Rel", "Jiang", "Lin", "TCSS"
#' @return score matrix
#' @export
#' @details The TCSS method requires `semData` generated by [godata()] with
#' `computeIC = TRUE` and `processTCSS = TRUE`.
#' @author Guangchuang Yu <https://yulab-smu.top>
termSim <- function(t1, t2, semData, method=c("Wang", "Resnik", "Rel", "Jiang", "Lin", "TCSS")) {

    method <- match.arg(method)

    if (all(is.na(t1)) || all(is.na(t2)))
        return(NA)

    t1 <- t1[!is.na(t1)]
    t2 <- t2[!is.na(t2)]

    ## genes in cluster may share go terms,
    ## and should affect similarity of clusters in clusterSim and mclusterSim.
    ##
    ## t1 <- unique(t1)
    ## t2 <- unique(t2)

    if ( method %in% c("Resnik", "Jiang", "Lin", "Rel") ) {
        return(infoContentMethod(t1, t2, method = method, semData))
    } else if (method == "Wang") {
        return(wangMethod(t1, t2, semData@ont))
    } else if ( method == "TCSS" ) {
        return(tcssMethod(t1, t2, semData))
    }
}

#' determine the topological cutoff for TCSS method
#'
#' @param OrgDb OrgDb object
#' @param keytype keytype
#' @param ont ontology : "BP", "MF", "CC"
#' @param combine_method one of "max", "BMA", "avg", "rcmax", "rcmax.avg"
#' @param ppidata A data.frame contains positive set and negative set.
#' Positive set is PPI pairs that already verified.
#' ppidata has three columns, column 1 and 2 are character, column 3
#' must be logical value:TRUE/FALSE.
#'
#' @return numeric, topological cutoff for TCSS subgraph construction. The
#' returned value can be passed to [godata()] via the `cutoff` argument together
#' with `processTCSS = TRUE`.
#' @export
#'
#' @examples
#' \dontrun{
#'     library(org.Hs.eg.db)
#'     library(STRINGdb)
#'
#'     string_db <- STRINGdb$new(version = "11.0", species = 9606,
#'     score_threshold = 700)
#'     string_proteins <- string_db$get_proteins()
#'
#'     #get relationship
#'     ppi <- string_db$get_interactions(string_proteins$protein_external_id)
#'
#'     ppi$from <- vapply(ppi$from, function(e)
#'                        strsplit(e, "9606.")[[1]][2], character(1))
#'     ppi$to <- vapply(ppi$to, function(e)
#'                        strsplit(e, "9606.")[[1]][2], character(1))
#'     len <- nrow(ppi)
#'
#'     #select length
#'     s_len <- 100
#'     pos_1 <- sample(len, s_len, replace = T)
#'     #negative set
#'     pos_2 <- sample(len, s_len, replace = T)
#'     pos_3 <- sample(len, s_len, replace = T)
#'     #union as ppidata
#'     ppidata <- data.frame(pro1 = c(ppi$from[pos_1], ppi$from[pos_2]),
#'      pro2 = c(ppi$to[pos_1], ppi$to[pos_3]),
#'      label = c(rep(TRUE, s_len), rep(FALSE, s_len)),
#'      stringsAsFactors = FALSE)
#'
#'     cutoff <- tcss_cutoff(OrgDb = org.Hs.eg.db, keytype = "ENSEMBLPROT",
#'     ont = "BP", combine_method = "max", ppidata)
#'
#'     semData <- godata(annoDb = org.Hs.eg.db, keytype = "ENSEMBLPROT",
#'     ont = "BP", computeIC = TRUE, processTCSS = TRUE, cutoff = cutoff)
#' }
tcss_cutoff <- function(OrgDb = NULL, keytype = "ENTREZID", ont,
                        combine_method = "max", ppidata) {

  combine_method <- match.arg(combine_method,
                              c("max", "BMA", "avg", "rcmax", "rcmax.avg"))

  semdata <- godata(OrgDb, keytype = keytype, ont = ont, computeIC = TRUE,
                    processTCSS = FALSE, cutoff = NULL)
  #cutoff is in the range of ICT value
  IC <- semdata@IC
  GO <- names(IC[!is.infinite(IC)])
  offspring <- getOffsprings(ont) 
  
  #compute ICT value for each term
  ICT <- computeICT(GO, offspring)
  #cutoffs, all possible cutoff values
  cutoffs <- seq(0.1, max(ICT) + 0.1, by = 0.1)
  #all genes/proteins that have none-zero annotations
  all_pro <- unique(semdata@geneAnno[, keytype])
  #filter the ppidata
  filtered_ppidata <- create_filtered_ppidata(all_pro, ppidata = ppidata)
  #calculate the similarity value for filtered_ppidata
  predict_result <- lapply(cutoffs, computePre,
                           filtered_ppidata = filtered_ppidata,
                           semdata = semdata,
                           combine_method = combine_method)

  #calculate the auc and F1_score
  auc_F1_score <- calc_auc_F1_score(predict_result,
                                    filtered_ppidata = filtered_ppidata)
  #decide the most appropriate cutoff
  decide_cutoff(auc_F1_score, cutoffs = cutoffs)
}

#' keep the proteins with none-zero annotations
#'
#' @param all_pro all proteins that have none-zero annotations
#' @param ppidata data.frame, already verified PPI data
#' @importFrom stats na.omit
#'
#' @return data.frame, annotated protein pairs and their labels
#' @noRd
create_filtered_ppidata <- function(all_pro, ppidata) {
  #check data type
  if (!(is.character(ppidata[, 1]) && is.character(ppidata[, 2]) &&
        is.logical(ppidata[, 3]))) {
    stop("ppidata must be a data.frame with three columns:character, character, logical")
  }

  ppidata <- na.omit(ppidata)

  #remove proteins that have zero annotations
  len1 <- ppidata[, 1] %in% all_pro
  len2 <- ppidata[, 2] %in% all_pro
  ppidata_exist <- ppidata[len1 & len2, ]
  filtered_ppidata <- unique(ppidata_exist)

  len <- dim(filtered_ppidata)[1]

  if (len == 0) {
    stop("filtered ppidata is empty, none items have GO annotation. Please input more data.")
  }

  nTrue <- sum(filtered_ppidata[, 3])
  nFalse <- len - nTrue

  if (nTrue == len || nFalse == len) {
    stop("The filtered ppidata lacks the necessary label:TRUE and FALSE. Please input more data.")
  }

  message(paste("positive set has", nTrue,
                "PPI pairs, negative set has", nFalse, "PPI pairs"))

  return(filtered_ppidata)
}

#' compute prediction value on filtered ppidata
#'
#' @param cutoff numeric, topological cutoff
#' @param filtered_ppidata data.frame, annotated protein pairs and their labels
#' @param semdata GOSemSimDATA object
#' @param combine_method "max" "BMA", "avg", "rcmax", "rcmax.avg"
#' @return numeric, the prediction value for the input cutoff
#' @noRd
#'
computePre <- function(cutoff, filtered_ppidata, semdata,
                       combine_method) {
  #tcssdata is updated with this input cutoff
  tcssdata <- process_tcss(semdata@ont, semdata@IC, cutoff = cutoff)

  semdata@tcssdata <- tcssdata
  #similarity value is calculated with the semdata
  vapply(seq_len(nrow(filtered_ppidata)), function(i) {
    res <- geneSim(filtered_ppidata[i, 1], filtered_ppidata[i, 2],
                   semData = semdata, measure = "TCSS",
                   combine = combine_method, drop = FALSE)
    if (is.list(res)) return(as.numeric(res[["geneSim"]]))
    as.numeric(res)
  }, numeric(1))
}

#' calculate auc and F1-score
#'
#' @param predict_result list, prediction value for all cutoffs
#' @param filtered_ppidata data.frame, annotated protein pairs and their labels
#' @return data.frame, auc and F1-score value for different cutoffs
#' @importFrom methods slot
#' @importFrom rlang check_installed
#' @noRd
#'
calc_auc_F1_score <- function(predict_result, filtered_ppidata) {
  # checking whether package 'ROCR' being installed
  check_installed('ROCR', 'for`calc_auc_F1_score()`.')
  # the label for PPIs, TRUE/FALSE
  label <- filtered_ppidata[, 3]
  pre_value <- lapply(predict_result, as.numeric)

  # prediction object
  pred <- lapply(pre_value, function(e) {
    pos_stay <- !is.na(e)
    ROCR::prediction(e[pos_stay], label[pos_stay],
                     label.ordering = c(FALSE, TRUE)
    )
  })
  # performance object for auc
  perf_auc <- lapply(pred, ROCR::performance, measure = "auc")
  # auc value
  auc <- unlist(lapply(perf_auc, function(e) slot(e, "y.values")[[1]]))
  # performance object for F1-score
  perf_F1_score <- lapply(pred, ROCR::performance, measure = "f")
  # F1-score value, average value at different semantic similarity cutoffs
  F1_score <- unlist(lapply(perf_F1_score, function(e) {
    mean(slot(e, "y.values")[[1]], na.rm = TRUE)
  }))

  #return as data.frame
  return(data.frame(auc = auc,
                    F1_score = F1_score,
                    stringsAsFactors = F))
}

#' select the most appropriate cutoff
#'
#' @param auc_F1_score data.frame, auc and F1-score value for different cutoffs
#' @param cutoffs vector, all possible cutoff values
#' @return vector, topological cutoff for given parameters
#' @noRd
#'
decide_cutoff <- function(auc_F1_score, cutoffs) {
  #product value satisfies the "both maximized" requirement
  auc_mutiply_F1 <- auc_F1_score[, "auc"] * auc_F1_score[, "F1_score"]
  #get the max product value
  pos <- which(auc_mutiply_F1 == max(auc_mutiply_F1))

  if (length(pos) == 1)  return(cutoffs[pos])

  #if not only one pair of auc and F1-score have same product
  #take the one with larger auc
  select_auc <- auc_F1_score[pos, "auc"]

  auc_pos <- which(select_auc == max(select_auc))

  if (length(auc_pos) == 1) return(cutoffs[pos[auc_pos]])

  #if more than one pair of auc and F1-score are both same
  #take the smaller cutoff for time saving
  return(cutoffs[pos[min(auc_pos)]])
}

library(GOSemSim)

context("TCSS")

test_that("TCSS self similarity is one", {
    hsGO <- godata(annoDb = "org.Hs.eg.db", ont = "BP", computeIC = TRUE, processTCSS = TRUE)
    x <- termSim("GO:0009987", "GO:0009987", hsGO, method = "TCSS")
    expect_equal(unname(x[1, 1]), 1)
})

test_that("TCSS handles invalid terms and missing tcssdata", {
    hsGO <- godata(annoDb = "org.Hs.eg.db", ont = "BP", computeIC = TRUE, processTCSS = TRUE)
    x <- termSim("BAD", "GO:0009987", hsGO, method = "TCSS")
    expect_true(is.na(x[1, 1]))

    hsGO_no_tcss <- godata(annoDb = "org.Hs.eg.db", ont = "BP", computeIC = TRUE)
    expect_error(
        termSim("GO:0009987", "GO:0009987", hsGO_no_tcss, method = "TCSS"),
        "tcssdata not found"
    )
})

test_that("TCSS ancestor cache is ontology-specific", {
    yulab.utils::initial_cache()
    bp_anc <- GOSemSim:::ancestors_envir("GO:0009987", "BP")
    mf_anc <- GOSemSim:::ancestors_envir("GO:0009987", "MF")

    expect_false(identical(bp_anc, mf_anc))
    expect_false(is.null(bp_anc))
    expect_true(is.null(mf_anc) || all(is.na(mf_anc)))
})

test_that("TCSS requires finite IC values", {
    expect_error(
        GOSemSim:::process_tcss("BP", numeric()),
        "IC data not found"
    )
    expect_error(
        GOSemSim:::process_tcss("BP", c("GO:test" = Inf)),
        "IC data not found"
    )
})

test_that("TCSS cutoff helpers validate and return numeric predictions", {
    expect_error(
        tcss_cutoff(ont = "BP", combine_method = "bad",
                    ppidata = data.frame(a = "1", b = "2", label = TRUE)),
        "'arg' should be one of"
    )

    expect_error(
        GOSemSim:::create_filtered_ppidata(
            all_pro = c("1", "2"),
            ppidata = data.frame(a = 1, b = "2", label = TRUE)
        ),
        "ppidata must be"
    )

    hsGO <- godata(annoDb = "org.Hs.eg.db", ont = "BP", computeIC = TRUE)
    genes <- unique(hsGO@geneAnno[, 1])
    ppidata <- data.frame(
        a = as.character(genes[1:2]),
        b = as.character(genes[2:3]),
        label = c(TRUE, FALSE),
        stringsAsFactors = FALSE
    )
    filtered <- GOSemSim:::create_filtered_ppidata(genes, ppidata)
    pred <- GOSemSim:::computePre(
        cutoff = 3.5,
        filtered_ppidata = filtered,
        semdata = hsGO,
        combine_method = "max"
    )

    expect_type(pred, "double")
    expect_length(pred, nrow(filtered))
})

test_that("TCSS cutoff AUC/F1 helper consumes numeric predictions", {
    testthat::skip_if_not_installed("ROCR")
    filtered <- data.frame(
        a = c("1", "2", "3", "4"),
        b = c("2", "3", "4", "5"),
        label = c(TRUE, FALSE, TRUE, FALSE),
        stringsAsFactors = FALSE
    )
    res <- GOSemSim:::calc_auc_F1_score(
        predict_result = list(c(0.9, 0.2, 0.8, NA),
                              c(0.7, 0.3, 0.6, 0.1)),
        filtered_ppidata = filtered
    )

    expect_s3_class(res, "data.frame")
    expect_named(res, c("auc", "F1_score"))
    expect_equal(nrow(res), 2)
})

##' parse GAF files
##'
##' given a GAF file, this function extracts the information from it
##' @title read.gaf
##' @rdname read-gaf
##' @param file GAF file
##' @param asis logical, whether output the original contains of the file and only works if 'add_indirect_GO = FALSE'
##' @param add_indirect_GO whether to add indirect GO annotation 
##' @return A data.frame. Original table if 'asis' works, otherwise contains 3 conlumns of 'GENE', 'GO' and 'ONTOLOGY'
##' @export
read.gaf <- function(file, asis = FALSE, add_indirect_GO = FALSE) {
  GafFile <- read.gaf2(file)
  if (!add_indirect_GO && asis) return(GafFile)

  new.data.frame <- GafFile[, c("DB_Object_ID", "GOID", "Aspect")]
  names(new.data.frame) <- c("GENE", "GO", "ONTOLOGY")
  ont <- setNames(c("MF", "CC", "BP"), c("F", "C", "P"))
  new.data.frame$ONTOLOGY <- ont[new.data.frame$ONTOLOGY]

  if (!add_indirect_GO) return(new.data.frame)

  ## use buildGOmap function to append indirect annotation
  buildGOmap(new.data.frame)
}

##' @importFrom GO.db GO.db
##' @importFrom AnnotationDbi columns
goid2term <- function(simplify = TRUE) {
  go.ALL <- AnnotationDbi::select(GO.db, keys(GO.db), columns(GO.db))
  if (simplify) {
    go.ALL <- go.ALL[, c("GOID", "TERM")]
  }

  return(go.ALL)
}

##' @rdname read-gaf
##' @export
parse_gff <- read.gaf

# only read the file with selected columns
##' @importFrom utils read.delim
read.gaf2 <- function(GafFile, nrows = -1) {
  cat("Reading ", GafFile, ": ", sep = "")
  GafFile <-
    read.delim(
      GafFile,
      sep = "\t",
      as.is = TRUE,
      quote = "\"",
      fill = TRUE,
      header = FALSE,
      nrows = nrows,
      comment.char = "!"
    )
  GafFile <- GafFile[, c(2, 3, 5, 7, 9, 10)]
  names(GafFile) <- c(
    "DB_Object_ID",
    "DB_Object_Symbol",
    "GOID",
    "Evidence_Code",
    "Aspect",
    "DB_Object_Name"
  )
  cat("found", nrow(GafFile), "rows in this GAF file")
  return(GafFile)
}

## ----style, echo=FALSE, results="asis", message=FALSE-------------------------
knitr::opts_chunk$set(tidy = FALSE,
                      warning = FALSE,
                      message = FALSE)

## ----echo=FALSE, results="hide", message=FALSE--------------------------------
Biocpkg <- function (pkg) {
    sprintf("[%s](http://bioconductor.org/packages/%s)", pkg, pkg)
}

#' Given two genes, calculate their semantic similarity and return the score with corresponding GO terms.
#'
#'
#' @title Semantic similarity between two genes
#' @param gene1 Entrez gene ID
#' @param gene2 Another Entrez gene ID
#' @template params-measure-combine
#' @param drop Evidence codes to drop; use `NULL` to keep all GO annotations
#' @return A list containing similarity value and corresponding GO terms
#' @seealso [goSim()] [mgoSim()] [mgeneSim()] [clusterSim()] [mclusterSim()]
#' @export
#' @examples
#' d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
#' geneSim("241", "251", semData = d, measure = "Wang")
#' @author Guangchuang Yu <https://yulab-smu.top>
geneSim <- function(gene1, gene2, semData, measure="Wang", drop="IEA", combine="BMA") {
    go1 <- gene2GO(gene1, semData, dropCodes = drop)
    go2 <- gene2GO(gene2, semData, dropCodes = drop)
    if (length(go1) == 0 || length(go2) == 0)
        return(NA)
    res <- mgoSim(go1, go2, semData = semData, measure = measure, combine = combine)
    return(list(geneSim = res, GO1 = go1, GO2 = go2))
}

#' Method TCSS for semantic similarity measuring
#'
#' @param t1 term vector
#' @param t2 term vector
#' @param semData GOSemSimDATA object
#'
#' @return vector, similarity score for t1 and t2
#' @noRd
#'
#' @examples
#' library(org.Hs.eg.db)
#' semdata <- godata(org.Hs.eg.db,
#'   keytype = "ENTREZID", ont = "BP",
#'   computeIC = TRUE, processTCSS = TRUE, cutoff = NULL
#' )
#' termSim("GO:0000012", "GO:0009987", semdata, method = "TCSS")
tcssMethod <- function(t1, t2, semData) {
    matrix(mapply(tcssMethod_internal,
                  rep(t1, length(t2)),
                  rep(t2, each = length(t1)),
                  MoreArgs = list(semData = semData)),
    dimnames = list(t1, t2), ncol = length(t2)
    )
}

#' process one term with one term
#'
#' @param ID1 term
#' @param ID2 term
#' @param semData GOSemSimDATA object
#'
#' @return numeric, similarity score for ID1 and ID2
#' @noRd
#'
tcssMethod_internal <- function(ID1, ID2, semData) {
    tcssdata <- semData@tcssdata
    ont <- semData@ont

    if (length(tcssdata) == 0) {
        stop("tcssdata not found, please re-generate your `semData` with `processTCSS = TRUE`...")
    }

    GOs <- names(tcssdata$clusid)

    if ((!ID1 %in% GOs) || (!ID2 %in% GOs)) return(NA)
    if (ID1 == ID2) return(1)

    # get common ancestors
    com_anc <- ancestors_in_common(ID1 = ID1, ID2 = ID2, ont = ont)
    com_anc <- com_anc[com_anc %in% GOs]

    if (length(com_anc) == 0) {
        return(NA)
    }

    sim_value <- lapply(com_anc, function(x) {
        vapply(tcssdata$ica[tcssdata$clusid[[x]]], function(y) y[x], numeric(1))
    })
    sim_value <- unlist(sim_value)

    ## # belonged cluster-ids for each ID
    ## clus1_list <- tcssdata[[ID1]][["clusid"]]
    ## clus2_list <- tcssdata[[ID2]][["clusid"]]

    ## # calculate within different clusters
    ## sim_value <- unlist(mapply(calc_lca,
    ##                            rep(clus1_list, length(clus2_list)),
    ##                            rep(clus2_list, each = length(clus1_list)),
    ##                            MoreArgs = list(
    ##                                ID1 = ID1, ID2 = ID2,
    ##                                tcssdata = tcssdata,
    ##                                com_anc = com_anc, ont = ont
    ##                            )))

    if (is.null(sim_value) || length(sim_value) == 0) {
        return(NA)
    }
    # here max value means lowest common ancestor
    max(sim_value)
}

## #' calculate lowest common ancestors' value
## #'
## #' @param clus1 character, cluster-id for ID1
## #' @param clus2 character, cluster-id for ID2
## #' @param ID1 term
## #' @param ID2 term
## #' @param tcssdata list, belonged clusters and its elements for all nodes
## #' @param com_anc character, common ancestors
## #' @param ont ontology
## #'
## #' @return numeric/NULL, similarity value for ID1 with clus1 and ID2 with clus2
## #' @noRd
## #'
## calc_lca <- function(clus1, clus2, ID1, ID2, tcssdata, com_anc, ont) {
##     # "meta" only represents meta-graph, do not have relations
##     if (clus1 == "meta" || clus2 == "meta") {
##         return(NULL)
##     }
##     # if the two clusters are the same one
##     if (identical(clus1, clus2)) {
##         # all cluster-nodes inside cluster
##         clus_content <- tcssdata[[clus1]]
##     } else {
##         # all cluster-nodes inside "meta" cluster
##         # common ancestors are from clus1 and clus2
##         clus_content <- tcssdata[["meta"]]
##         com_anc <- ancestors_in_common(ID1 = clus1, ID2 = clus2, ont = ont)
##     }

##     # get common ancestors' position, nomatch helps not to introduce NA
##     com_anc_pos <- match(com_anc, clus_content[["GO"]], nomatch = 0)
##     # common ancestors' ica value is all the possible sim value
##     sim_value <- clus_content[["ica"]][com_anc_pos]

##     if (is.null(sim_value) || length(sim_value) == 0) {
##         return(NULL)
##     }
##     # here max similarity value means one of lowest common ancestors
##     max(sim_value)
## }

#' collect common ancestors
#'
#' @param ID1 term
#' @param ID2 term
#' @param ont ontology
#'
#' @return character, common ancestors for ID1 and ID2
#' @noRd
#'
ancestors_in_common <- function(ID1, ID2, ont) {
    ancestor1 <- ancestors_envir(ID1, ont)
    ancestor2 <- ancestors_envir(ID2, ont)

    if (ID1 == ID2 || ID1 %in% ancestor2) {
        return(ID1)
    }

    if (ID2 %in% ancestor1) {
        return(ID2)
    }

    setdiff(intersect(ancestor1, ancestor2), "all")
}

#' get ancestors from environment
#'
#' @param ID term
#' @param ont ontology
#'
#' @return ancestors for ID
#' @noRd
ancestors_envir <- function(ID, ont) {
    cache_key <- paste(ID, ont, sep = "|")
    anc <- yulab.utils::get_cache_element("GOSemSim_ancCache", cache_key)
    if (!is.null(anc)) return(anc)
    ancestors <- getAncestors(ont)[[ID]]
    e <- stats::setNames(list(ancestors), cache_key)
    yulab.utils::update_cache_item("GOSemSim_ancCache", e)
    ancestors
}

gene2GO <- function(gene, godata, dropCodes) {
    goAnno <- godata@geneAnno
    if (! "EVIDENCE" %in% colnames(goAnno)) {
        warning("Evidence codes not found, 'drop' parameter will be ignored...")
    } else {
        goAnno <- goAnno[!goAnno$EVIDENCE %in% dropCodes,]
    }
    go <- as.character(unique(goAnno[goAnno[,1] == gene, "GO"]))
    go[!is.na(go)]
}

#' Semantic similarity between two gene clusters
#'
#'
#' @title Semantic similarity between two gene clusters
#' @param cluster1 A set of gene IDs
#' @param cluster2 Another set of gene IDs
#' @template params-measure-combine
#' @param drop Evidence codes to drop; use `NULL` to keep all GO annotations
#' @return similarity
#' @seealso [goSim()] [mgoSim()] [geneSim()] [mgeneSim()] [clusterSim()] [mclusterSim()]
#' @export
#' @examples
#' d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
#' cluster1 <- c("835", "5261", "241", "994")
#' cluster2 <- c("307", "308", "317", "321", "506", "540", "378", "388", "396")
#' clusterSim(cluster1, cluster2, semData = d, measure = "Wang")
#' @author Guangchuang Yu <https://yulab-smu.top>
clusterSim <- function(cluster1, cluster2, semData, measure="Wang", drop="IEA", combine="BMA") {
    cgo1 <- sapply(cluster1, gene2GO, semData, dropCodes = drop)
    cgo2 <- sapply(cluster2, gene2GO, semData, dropCodes = drop)
    cgo1 <- unlist(cgo1)
    cgo2 <- unlist(cgo2)
    res <- mgoSim(cgo1, cgo2, semData, measure = measure, combine = combine)
    return(res)
}

##'
##' given a BLAST2GO file, this function extracts the information from it and make it use for TERM2GENE.
##' @title read.blast2go
##' @param file BLAST2GO file
##' @param add_indirect_GO whether add indirect GO annotation 
##' @importFrom rlang check_installed
##' @return a data frame with three columns: GENE, GO and ONTOLOGY
##' @export
read.blast2go <- function(file, add_indirect_GO = FALSE) {
    check_installed("readr", 'for `read.blast2go()`.')
    check_installed('tidyr', 'for `read.blast2go()`.')
    check_installed('tidyselect', 'for `read.blaset2go()`.')

    # blast2go <- utils::read.table(file, header = TRUE, sep = "\t", stringsAsFactors = FALSE, fill = TRUE) # has bugs
    blast2go <- yulab.utils::yread(file, readr::read_delim)
    x <- blast2go[, c("Sequence Name", "GO Accession", "GO Domains")]
    names(x) <- c("GENE", "GO", "ONTOLOGY")
    x <- x[!is.na(x[["GO"]]), ]

    y <- tidyr::separate_rows(x, tidyselect::all_of(c("GO", "ONTOLOGY")), sep = ", ")
    y$GO <- sub("^\\s+", "", y$GO)
    y$ONTOLOGY <- sub("^\\s+", "", y$ONTOLOGY)
    
    ont <- setNames(c("MF", "CC", "BP"), c("F", "C", "P"))
    y$ONTOLOGY <- ont[y$ONTOLOGY]
    y <- as.data.frame(y)

    if (!add_indirect_GO) {
        return(y)
    }

    buildGOmap(y)
}

#' @keywords internal
buildGoMatrix <- function(uniqueGO, semData, measure) {
    mgoSim(uniqueGO, uniqueGO, semData, measure = measure, combine = NULL)
}

#' @keywords internal
uniqueGOFrom <- function(golist) {
    unique(unlist(golist))
}

#' @keywords internal
subsetCombine <- function(go_matrix, gos1, gos2, combine) {
    combineScores(go_matrix[gos1, gos2, drop = FALSE], combine = combine)
}

getOffspringIdx <- function(ont, goids) {
    key <- paste0("offspring_idx_", ont, "_", digest::digest(goids))
    res <- yulab.utils::get_cache_element("GOSemSim_offspring_idx", key)
    if (!is.null(res)) return(res)
    off <- getOffsprings(ont)
    pos <- stats::setNames(seq_along(goids), goids)
    idx <- lapply(goids, function(id) {
        ids <- off[[id]]
        if (is.null(ids)) integer(0) else as.integer(stats::na.omit(pos[ids]))
    })
    names(idx) <- goids
    e <- list()
    e[[key]] <- idx
    yulab.utils::update_cache_item("GOSemSim_offspring_idx", e)
    idx
}

#' combining similarity matrix to similarity score
#'
#' Functions for combining similarity matrix to similarity score
#'
#'
#' @param SimScores similarity matrix
#' @param combine combine method
#' @return similarity value
#' @export
#' @author Guangchuang Yu <https://yulab-smu.top>
combineScores <- function(SimScores, combine) {

    if (length(combine) == 0) {  #if not define combine
        return(round(SimScores, digits=3))
    }

    ## if combine was defined...
    if(!sum(!is.na(SimScores))) return (NA)

    if (is.vector(SimScores) || nrow(SimScores)==1 || ncol(SimScores)==1) {
        if (combine == "avg") {
            return(round(mean(SimScores, na.rm=TRUE), digits=3))
        } else {
            return (round(max(SimScores, na.rm=TRUE), digits=3))
        }
    }


    row.na.idx <- apply(SimScores, 1, function(i) all(is.na(i)))
    if (any(row.na.idx)) {
        SimScores <- SimScores[-which(row.na.idx), ]
    }

    if (! is.null(dim(SimScores)) ) {
        col.na.idx <- apply(SimScores, 2, function(i) all(is.na(i)))
        if (any(col.na.idx)) {
            SimScores <- SimScores[ , -which(col.na.idx)]
        }
    }
    if (is.vector(SimScores) || nrow(SimScores)==1 || ncol(SimScores)==1) {
        if (combine == "avg") {
            return(round(mean(SimScores, na.rm=TRUE), digits=3))
        } else {
            return (round(max(SimScores, na.rm=TRUE), digits=3))
        }
    }

    if (combine        == "avg") {
        result   <- mean(SimScores, na.rm=TRUE)
    } else if (combine == "max") {
        result   <- max(SimScores, na.rm=TRUE)
    } else if (combine == "rcmax") {
        rowScore <- mean(apply(SimScores, 1, max, na.rm=TRUE))
        colScore <- mean(apply(SimScores, 2, max, na.rm=TRUE))
        result   <- max(rowScore, colScore)
    } else if (combine == "rcmax.avg" || combine == "BMA") {
        result   <- sum( apply(SimScores, 1, max, na.rm=TRUE),
                        apply(SimScores, 2, max, na.rm=TRUE)
                        ) / sum(dim(SimScores))
    }

    return (round(result, digits=3))
}

#' Information Content Based Methods for semantic similarity measuring
#'
#' implemented for methods proposed by Resnik, Jiang, Lin and Schlicker.
#' @title information content based methods
#' @param ID1 Ontology Term
#' @param ID2 Ontology Term
#' @param method one of "Resnik", "Jiang", "Lin" and "Rel", "TCSS".
#' @param godata GOSemSimDATA object
#' @return semantic similarity score
#' @useDynLib GOSemSim
#' @author Guangchuang Yu <https://yulab-smu.top>
infoContentMethod <- function(ID1,
                              ID2,
                              method,
                              godata) {
    ## IC is biased
    ## because the IC of a term is dependent of its children but not on its parents.
    ont <- godata@ont
    IC <- godata@IC

    if (length(IC) == 0) {
        stop("IC data not found, please re-generate your `semData` with `computeIC=TRUE`...")
    }

    if (ont %in% c("MF", "BP", "CC", "HDO", "MPO", "HPO")) {
        ## .anc <- tryCatch(getAncestors(ont)[union(ID1,ID2)], error=function(e) NULL)
        ## if (is.null(.anc)) {
        ##     ## https://support.bioconductor.org/p/105822/
        ##     return(NA)
        ## }
        ## .anc <- AnnotationDbi::as.list(.anc)

        ## if some IDs are not valid, the above code will leading to return NA directly.
        .anc <- getAncestors(ont)
        allid <- union(ID1, ID2)
        .anc <- .anc[allid]
        .anc <- .anc[!vapply(.anc, is.empty, logical(1))]

        ## invalid_ids <- c(ID1[!ID1 %in% names(.anc)], ID2[!ID2 %in% names(.anc)])
        ## if (length(invalid_ids) > 0) {
        ##     message("The following IDs are not valid and will be removed:", paste(invalid_ids, collapse=","))
        ##     # allid <- allid[!allid %in% invalid_ids]
        ## }
    } else {
        mesh_getAnc <- eval(parse(text="meshes:::getAncestors"))
        .anc <- lapply(union(ID1, ID2), mesh_getAnc)
        names(.anc) <- union(ID1, ID2)
    }
    return ( infoContentMethod_cpp( ID1, ID2,
                 .anc, IC,
                 method, ont ) )
}

is.empty <- function(x) {
    if (is.null(x)) return(TRUE)
    if (all(is.na(x))) return(TRUE)
    return(FALSE)
}


## infoContentMethod <- function(ID1,
##                               ID2,
##                               ont="DO",
##                               method,
##                               organism="human") {
##     IC <- getIC(organism, ont)

##     ## more specific term, larger IC value.
##     ## Normalized, all divide the most informative IC.
##     ## all IC values range from 0(root node) to 1(most specific node)
##     mic <- max(IC[IC!=Inf])

##     if (ont == "DO") {
##         topNode <- "DOID:4"
##     } else {
##         topNode <- "all"
##     }

##     IC[topNode] = 0

##     ic1 <- IC[ID1]/mic
##     ic2 <- IC[ID2]/mic

##     if (ic1 == 0 || ic2 == 0)
##         return (NA)

##     ancestor1 <- getAncestors(ont)[[ID1]]
##     ancestor2 <- getAncestors(ont)[[ID2]]
##     if (ID1 == ID2) {
##         commonAncestor <- ID1
##     } else if (ID1 %in% ancestor2) {
##         commonAncestor <- ID1
##     } else if (ID2 %in% ancestor1) {
##         commonAncestor <- ID2
##     } else {
##         commonAncestor <- intersect(ancestor1, ancestor2)
##     }
##     if (length(commonAncestor) == 0) return (NA)

##     ##Information Content of the most informative common ancestor (MICA)
##     mica <- max(IC[commonAncestor])/mic

##     ## IC is biased
##     ## because the IC of a term is dependent of its children but not on its parents.
##     sim <- switch(method,
##                   Resnik = mica, ## Resnik does not consider how distant the terms are from their common ancestor.
##                   ## Lin and Jiang take that distance into account.
##                   Lin = 2*mica/(ic1+ic2),
##                   Jiang = 1 - min(1, -2*mica + ic1 + ic2),
##                   Rel = 2*mica/(ic1+ic2)*(1-exp(-mica*mic))  ## mica*mic equals to the original IC value. and exp(-mica*mic) equals to the probability of the term's occurence.
##                   )
##     return (sim)
## }

.initial <- function() {
    e <- new.env(parent = emptyenv())
    gotbl <- tryCatch(utils::data("gotbl", package="GOSemSim", envir=e),
                      error = function(e) NULL)
    if (is.null(gotbl)) return(NULL)
    
    yulab.utils::update_cache_item(".GOSemSimEnv", list(gotbl = e$gotbl))
}

ensure_gotbl_cached <- function() {
    gotbl <- yulab.utils::get_cache_element(".GOSemSimEnv", "gotbl")
    if (is.null(gotbl) || !is.data.frame(gotbl)) {
        .initial()
        gotbl <- yulab.utils::get_cache_element(".GOSemSimEnv", "gotbl")
    }
    return(gotbl)
}

get_gosemsim_env <- function() {
    ensure_gotbl_cached()
    invisible(NULL)
}

supported_GO <- function() c("BP", "CC", "MF")
supported_DO <- function() c("DO", "HDO", "HPO", "MPO")

is_supported_go <- function(ont) ont %in% supported_GO()
is_supported_do <- function(ont) ont %in% supported_DO()


##' @importFrom GO.db GOMFANCESTOR
##' @importFrom GO.db GOBPANCESTOR
##' @importFrom GO.db GOCCANCESTOR
getAncestors <- function(ont) {
    if (is_supported_go(ont)) {
        Ancestors <- switch(ont,
                            MF = GOMFANCESTOR,
                            BP = GOBPANCESTOR,
                            CC = GOCCANCESTOR
                            )
        anc <- AnnotationDbi::as.list(Ancestors)
        return(anc)
    }

    get_onto_data(ont, output = 'list', 'ancestor') 
}

##' @importFrom GO.db GOMFPARENTS
##' @importFrom GO.db GOBPPARENTS
##' @importFrom GO.db GOCCPARENTS
getParents <- function(ont) {
    if (is_supported_go(ont)) {
        Parents <- switch(ont,
                        MF = GOMFPARENTS,
                        BP = GOBPPARENTS,
                        CC = GOCCPARENTS
                        )
        parent <- AnnotationDbi::as.list(Parents)
        return(parent)
    }

    get_onto_data(ont, output = 'list', 'parent') 
}

##' @importFrom GO.db GOMFOFFSPRING
##' @importFrom GO.db GOBPOFFSPRING
##' @importFrom GO.db GOCCOFFSPRING
getOffsprings <- function(ont) {
    if (is_supported_go(ont)) {
        Offsprings <- switch(ont,
                        MF = GOMFOFFSPRING,
                        BP = GOBPOFFSPRING,
                        CC = GOCCOFFSPRING
                        )
        offspring <- AnnotationDbi::as.list(Offsprings)
        return(offspring)
    }

    get_onto_data(ont, output = 'list', 'offspring') 
}

##' @importFrom GO.db GOTERM
##' @importFrom AnnotationDbi toTable
prepare_relation_df <- function() {
    gtb <- toTable(GOTERM)
    gtb <- gtb[,c(2:4)]
    gtb <- unique(gtb)
    
    ptb <- lapply(c("BP", "MF", "CC"), function(ont) {
        id <- with(gtb, go_id[Ontology == ont])
        parentMap <- getParents(ont)
        # pid <- AnnotationDbi::mget(id, parentMap)
        pid <- parentMap[id]

        n <- sapply(pid, length)
        cid <- rep(names(pid), times=n)
        relationship <- unlist(lapply(pid, names))
        
        data.frame(id=cid,
                   relationship=relationship,
                   parent=unlist(pid),
                   stringsAsFactors = FALSE)
    }) 
    ptb <- do.call('rbind', ptb)

    gotbl <- merge(gtb, ptb, by.x="go_id", by.y="id")
    save(gotbl, file="gotbl.rda", compress="xz")
    invisible(gotbl)
}

#' @title Get organism name from OrgDb object
#' @param object OrgDb object or OrgDb package name
#' @return Organism name
#' @importFrom yulab.utils load_OrgDb
#' @export
#' @author Guangchuang Yu
get_organism <- function(object) {
    OrgDb <- load_OrgDb(object)
    AnnotationDbi::species(OrgDb)
}

#' @importFrom yulab.utils load_OrgDb
#' @export
yulab.utils::load_OrgDb

library(GOSemSim)
library(GO.db)

context("GO")

test_that("parent node", {
    goid <- 'GO:0004022'
    x <- as.list(GOSemSim:::getParents('MF')[goid])
    expect_equal(names(x),goid)

    pid <- x[[1]]
    expect_true( goid %in% as.list(GOMFCHILDREN[pid])[[1]] )
})

##' @importClassesFrom AnnotationDbi AnnotationDb
##' @importFrom methods setRefClass
setRefClass("OntDb", contains="AnnotationDb")

#' @importMethodsFrom AnnotationDbi keys
#' @importMethodsFrom AnnotationDbi toTable
setMethod("keys", "OntDb",
    function(x, keytype, ...){
        if(missing(keytype)) keytype <- "id"
        term <- toTable(x)
        term[, keytype]
    }
)

#' @importMethodsFrom AnnotationDbi keytypes
setMethod("keytypes", "OntDb",
    function(x) {
        c("id", "term")
    }

)


#' @importFrom DBI dbReadTable
setMethod("toTable", "OntDb",
    function(x) {
        setNames(dbReadTable(dbconn(x), 'term'), c("id", "term"))
    }
)



#' @importMethodsFrom AnnotationDbi select
#' @importMethodsFrom AnnotationDbi dbconn
setMethod("select", "OntDb",
    function(x, keys, columns, keytype, ...){
        if (missing(keytype)) keytype <- "id"
        keytype <- match.arg(keytype, c("id","term"))
        strKeys <- paste0("\"", keys, "\"", collapse = ",")
        if (keytype == "term") {
            sql_key <- paste("SELECT doid FROM do_term WHERE term in (",
                strKeys, ")")
            doids <- dbQuery(dbconn(x), sql_key)[, 1]
            strKeys <- paste0("\"", doids, "\"", collapse = ",")
        }
        columns <- unique(c("id", columns))

        sqls <- paste("SELECT ", paste(columns, collapse = ","),
            " FROM term")
        columns2 <- setdiff(columns, c("id", "term"))
        for (col in columns2) {
            leftJoin <- paste0("LEFT JOIN  ", col, " USING (id)")
            sqls <- c(sqls, leftJoin)
        }
        sqls <- c(sqls, paste0("WHERE term.id in (", strKeys, ")"))
        sqls <- paste(sqls, collapse = " ")
        res <- dbQuery(dbconn(x), sqls)
        res
    }
)

dbQuery <- getFromNamespace("dbQuery", "AnnotationDbi")

#' @importMethodsFrom AnnotationDbi columns
setMethod("columns", "OntDb",
    function(x) {
        c("id","term", "alias", "synonym", "parent", "children",
            "ancestor", "offspring")
    }
)


get_onto_data <- function(ont = "HDO", output='list', table="offspring") {
    x <- load_onto(ont)
    output <- match.arg(output, c("data.frame", "list"))
    res <- dbReadTable(dbconn(x), table)
    if (output == 'data.frame') return(res)

    # column 1 is ID, column 2 is the related term
    split(res[,2], res[,1]) 
}

#' Load Ontology Database
#' 
#' @param onto character. The ontology to load (e.g., "HDO").
#' @return An `AnnotationDb` object.
#' @importFrom digest digest
#' @importFrom AnnotationDbi loadDb
#' @importFrom yulab.utils download_yulab_file
#' @importFrom yulab.utils user_dir
#' @keywords internal
load_onto <- function(onto = "HDO") {
    .onto <- sprintf(".onto_%s", onto)
    
    db <- yulab.utils::get_cache_element(".GOSemSimEnv", .onto)
    if (!is.null(db)) return(db)

    dbfile <- sprintf("%s.sqlite", onto)
    urls <- c("https://yulab-smu.top/DOSE",
              "https://raw.githubusercontent.com/YuLab-SMU/DOSE/refs/heads/gh-pages")

    # use download_yulab_file from yulab.utils to handle multiple mirrors
    dbfile <- download_yulab_file(dbfile, urls, gzfile = TRUE, appname = "GOSemSim")

    db <- loadDb(dbfile)
    
    yulab.utils::update_cache_item(".GOSemSimEnv", setNames(list(db), .onto))
    return(db)
}

#' Given two sets of GO terms, calculate their semantic similarity.
#'
#'
#' @title Semantic similarity between two GO term sets
#' @param GO1 A set of GO terms
#' @param GO2 Another set of GO terms
#' @template params-measure-combine
#' @return similarity
#' @seealso [goSim()] [mgoSim()] [geneSim()] [mgeneSim()] [clusterSim()] [mclusterSim()]
#' @export
#' @examples
#' d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
#' go1 <- c("GO:0004022", "GO:0004024", "GO:0004023")
#' go2 <- c("GO:0009055", "GO:0020037")
#' mgoSim("GO:0003824", go2, semData = d, measure = "Wang")
#' mgoSim(go1, go2, semData = d, measure = "Wang")
#' @author Guangchuang Yu <https://yulab-smu.top>
mgoSim <- function(GO1, GO2, semData, measure="Wang", combine="BMA") {
    scores <- termSim(GO1, GO2, semData, method = measure)
    res <- combineScores(scores, combine)
    return(round(res, digits = 3))
}

#' Calculate pairwise semantic similarities for a given list of genes.
#'
#'
#' @title Pairwise semantic similarity for a list of genes
#' @param genes A list of Entrez gene IDs
#' @template params-measure-combine
#' @param drop Evidence codes to drop; use `NULL` to keep all GO annotations
#' @param verbose Whether to show a progress bar
#' @return similarity matrix
#' @seealso [goSim()] [mgoSim()] [geneSim()] [mgeneSim()] [clusterSim()] [mclusterSim()]
#' @export
#' @importFrom utils setTxtProgressBar txtProgressBar
#' @examples
#' d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
#' mgeneSim(c("835", "5261", "241"), semData = d, measure = "Wang")
#' @author Guangchuang Yu <https://yulab-smu.top>
mgeneSim <- function(genes, semData, measure="Wang", drop="IEA", combine="BMA", verbose=TRUE) {
    genes <- unique(as.character(genes))
    n <- length(genes)
    scores <- matrix(NA, nrow=n, ncol=n)
    rownames(scores) <- genes
    colnames(scores) <- genes

    gos <- lapply(genes, gene2GO, godata = semData, dropCodes = drop)
    uniqueGO <- uniqueGOFrom(gos)
    go_matrix <- buildGoMatrix(uniqueGO, semData, measure)
    if (verbose) {
      cnt <- 1
      pb <- txtProgressBar(min=0, max=sum(1:n), style=3)
    }
    for (i in seq_along(genes)) {
        for (j in seq_len(i)){
            if (verbose) {
                setTxtProgressBar(pb, cnt)
                cnt <- cnt + 1
            }
            scores[i, j] <- subsetCombine(go_matrix, gos[[i]], gos[[j]], combine)
            scores[j, i] <- scores[i, j]
        }
    }
    if (verbose)
        close(pb)
    removeRowNA <- apply(!is.na(scores), 1, sum) > 0
    removeColNA <- apply(!is.na(scores), 2, sum) > 0
    return(scores[removeRowNA, removeColNA, drop = FALSE])
}

#' Addding indirect GO annotation
#'
#' provided by a data.frame of GO TERM (column 1), GENE (column 2) and ONTOLOGY (optional) that
#' describes GO direct annotation,
#' this function will add indirect GO annotation of genes.
#'
#' @title buildGOmap
#' @param TERM2GENE data.frame with two or three columns of GO TERM, GENE and ONTOLOGY (optional)
#' @return data.frame, GO annotation with direct and indirect annotation
#' @importMethodsFrom AnnotationDbi as.list
#' @importFrom GO.db GOMFANCESTOR
#' @importFrom GO.db GOBPANCESTOR
#' @importFrom GO.db GOCCANCESTOR
#' @export
#' @author Yu Guangchuang
buildGOmap <- function(TERM2GENE) {
    mfanc <- as.list(GOMFANCESTOR)
    ccanc <- as.list(GOCCANCESTOR)
    bpanc <- as.list(GOBPANCESTOR)

    if (!'ONTOLOGY' %in% names(TERM2GENE)) {
        anc <- c(mfanc, ccanc, bpanc)
        res <- buildGOmap_internal(TERM2GENE, anc)
        return(res)
    }

    anc <- list(MF=mfanc, CC=ccanc, BP=bpanc)
    y <- split(TERM2GENE, TERM2GENE$ONTOLOGY)

    res <- lapply(names(y), function(i) {
        d <- buildGOmap_internal(y[[i]], anc[[i]])
        d$ONTOLOGY <- i
        return(d)
    }) |> do.call('rbind', args = _)

    return(res)
}

##' @importFrom stats setNames
##' @importFrom yulab.utils ls2df
buildGOmap_internal <- function(TERM2GENE, anc) {
    res <- setNames(anc[TERM2GENE[,1]], TERM2GENE[,2]) |> 
        ls2df() |>
        unique()

    res <- setNames(res[, c(2,1)], names(TERM2GENE)[1:2])
    res <- res[res[,1] != "all", ]
    res <- rbind(TERM2GENE[,1:2], res)
    return(res)
}

# old and slow version
##' @importMethodsFrom AnnotationDbi mget
##' @importFrom utils stack
buildGOmap2 <- function(gomap) {

    ## remove empty GO annotation
    gomap <- gomap[gomap[,1] != "", ]
    
    Gene2GO <- split(as.character(gomap[,1]), as.character(gomap[,2]))

    Gene2ALLGO <- lapply(Gene2GO,
                         function(i) {
                             mfans <- unlist(mget(i, GOMFANCESTOR, ifnotfound=NA))
                             bpans <- unlist(mget(i, GOBPANCESTOR, ifnotfound=NA))
                             ccans <- unlist(mget(i, GOCCANCESTOR, ifnotfound=NA))
                             ans <- c(mfans, bpans, ccans)
                             ans <- ans[ !is.na(ans) ]
                             ans <- c(i, ans)
                             ans <- unique(ans)
                             ans <- ans[ans != "all"]
                             return(ans)
                         })

    ## AMF <- as.list(GOMFANCESTOR)
    ## ACC <- as.list(GOCCANCESTOR)
    ## ABP <- as.list(GOBPANCESTOR)

    ## Gene2ALLGO <- lapply(Gene2GO, function(i) {
    ##     mfans <- AMF[i]
    ##     bpans <- ABP[i]
    ##     ccans <- ACC[i]
    ##     ans <- unlist(c(mfans,  bpans,  ccans))
    ##     ans <- ans[ !is.na(ans) ]
    ##     ans <- c(i, ans)
    ##     ans <- unique(ans)
    ##     ans <- ans[ans != "all"]
    ##     return(ans)
    ## })

    go2gene <- stack(Gene2ALLGO)
    colnames(go2gene) <- c("GO", "Gene")
    
    return(go2gene)
}

computeIC <- function(goAnno, ont) {
    ## goAnno, see godata function
    godata <- ensure_gotbl_cached()
    
    goids <- unique(godata[godata$Ontology == ont, "go_id"])
    ## all GO terms appearing in an given ontology ###########
    goterms=goAnno$GO
    gocount <- table(goterms)
    ## goid of specific organism and selected category.
    goname  <- names(gocount) 

    ## ensure goterms not appearing in the specific annotation have 0 frequency..
    go.diff        <- setdiff(goids, goname)
    m              <- double(length(go.diff))
    names(m)       <- go.diff
    gocount        <- as.vector(gocount)
    names(gocount) <- goname
    gocount        <- c(gocount, m)

    offspring_idx <- getOffspringIdx(ont, goids)
    gc <- gocount[goids]
    desc_sum <- vapply(goids, function(id) {
        ids <- offspring_idx[[id]]
        if (length(ids) == 0) 0 else sum(gc[ids], na.rm = TRUE)
    }, numeric(1))
    cnt <- gc + desc_sum
    names(cnt) <- goids
    
    ## the probabilities of occurrence of GO terms in a specific corpus.
    p <- cnt/sum(gocount)
    ## IC of GO terms was quantified as the negative log likelihood.
    IC <- -log(p)
    return(IC)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

infoContentMethod_cpp <- function(id1_, id2_, anc_, ic_, method_, ont_) {
    .Call('_GOSemSim_infoContentMethod_cpp', PACKAGE = 'GOSemSim', id1_, id2_, anc_, ic_, method_, ont_)
}

#' Calculate pairwise semantic similarities for a list of gene clusters.
#'
#'
#' @title Pairwise semantic similarities for a list of gene clusters
#' @param clusters A list of gene clusters
#' @template params-measure-combine
#' @param drop Evidence codes to drop; use `NULL` to keep all GO annotations
#' @return similarity matrix
#' @seealso [goSim()] [mgoSim()] [geneSim()] [mgeneSim()] [clusterSim()] [mclusterSim()]
#' @export
#' @examples
#' d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
#' cluster1 <- c("835", "5261", "241")
#' cluster2 <- c("578", "582")
#' cluster3 <- c("307", "308", "317")
#' clusters <- list(a = cluster1, b = cluster2, c = cluster3)
#' mclusterSim(clusters, semData = d, measure = "Wang")
#' @author Guangchuang Yu <https://yulab-smu.top>
mclusterSim <- function(clusters, semData, measure="Wang", drop="IEA", combine="BMA") {
    n <- length(clusters)
    cluster_gos <- list()
    for (i in 1:n) {
        cluster_gos[[i]] <- sapply(clusters[[i]], gene2GO, semData, dropCodes = drop)
    }

    uniqueGO <- uniqueGOFrom(cluster_gos)
    go_matrix <- buildGoMatrix(uniqueGO, semData, measure)

    scores <- matrix(NA, nrow=n, ncol=n)
    rownames(scores) <- names(clusters)
    colnames(scores) <- names(clusters)

    for (i in seq_along(clusters)) {
        gos1 <- unlist(cluster_gos[[i]])
        gos1 <- gos1[!is.na(gos1)]
        for (j in seq_len(i)) {
            gos2 <- unlist(cluster_gos[[j]])
            gos2 <- gos2[!is.na(gos2)]
            if (length(gos1) != 0 && length(gos2) !=0)
                scores[i, j] <- subsetCombine(go_matrix, gos1, gos2, combine)
                scores[j, i] <- scores[i, j]
        }
    }
    removeRowNA <- apply(!is.na(scores), 1, sum) > 0
    removeColNA <- apply(!is.na(scores), 2, sum) > 0
    return(scores[removeRowNA, removeColNA, drop = FALSE])
}

#' prepare GO DATA for measuring semantic similarity
#'
#'
#' @title godata
#' @param OrgDb OrgDb object (will be removed in future, please use annoDb instead)
#' @param annoDb GO annotation database,
#' can be OrgDb or a data.frame contains three columns of 'GENE', 'GO' and 'ONTOLOGY'.
#' @param keytype keytype
#' @param ont one of 'BP', 'MF', 'CC'
#' @param computeIC logical, whether computer IC
#' @param processTCSS logical, whether to prepare TCSS data. TCSS requires
#' `computeIC = TRUE`, which is enabled automatically when `processTCSS = TRUE`.
#' @param cutoff topology cutoff for TCSS subgraph construction. If `NULL`,
#' empirical defaults are used; users can estimate a cutoff with [tcss_cutoff()].
#' @return GOSemSimDATA object
#' @export
#' @importFrom methods new
#' @author Guangchuang Yu
godata <- function(OrgDb=NULL, annoDb=NULL, keytype = "ENTREZID",
                   ont, computeIC = TRUE, processTCSS = FALSE, cutoff = NULL) {
    if (processTCSS) computeIC <- TRUE

    ont <- toupper(ont)
    ont <- match.arg(ont, supported_GO())

    if (is.null(OrgDb) && is.null(annoDb)) {
        return(new("GOSemSimDATA",
        ont = ont
        ))
    }

    if (!is.null(OrgDb)) {
      warning("use 'annoDb' instead of 'OrgDb'")
      annoDb <- OrgDb
    }
    if (is.character(annoDb)) {
      annoDb <- load_OrgDb(annoDb) 
    }

    md <- data.frame()
    if (inherits(annoDb, 'OrgDb')) {
      goAnno <- parse_orgDb(annoDb, keytype)
      md <- AnnotationDbi::metadata(annoDb)
    } else if (inherits(annoDb, 'gson')) {
      ## to be supported
    } else { # for data.frame
      goAnno <- check_goAnno(annoDb)
    }

    goAnno <- goAnno[goAnno$ONTOLOGY == ont, ]
    if (computeIC) {
        message("preparing IC data...")
        IC <- computeIC(goAnno, ont)
        if (processTCSS) {
            message("preparing TCSS data...")
            tcssdata <- process_tcss(ont, IC = IC, cutoff = cutoff)
        }
    }

    res <- new("GOSemSimDATA",
      keys = unique(goAnno[,1]),
      ont = ont,
      geneAnno = goAnno,
      metadata = md
    )
    if (computeIC) {
        res@IC <- IC
        if (processTCSS) {
            res@tcssdata <- tcssdata
        }
    }

    return(res)
}

check_goAnno <- function(goAnno) {
  # check whether the data frame contains neccessary columns.
  if (!is.data.frame(goAnno)) {
      stop("annoDb should be a data.frame (or OrgDb object).")
  }

  ## suppose 1st column is GENE ID and should contains GO and ONTOLOGY columns
  ## maybe we should force names(goAnno)[1] == "GENE"
  if (!all(c("GO", "ONTOLOGY") %in% names(goAnno))) {
    stop("annoDb as a data.frame should contains 'GO' and 'ONTOLOGY' columns.")
  }

  if (ncol(goAnno) < 3) {
      warning("annoDb as a data.frame should have at least 3 columns: GENE, GO and ONTOLOGY. Assuming the first column is GENE ID.")
  }

  return(goAnno)
}

parse_orgDb <- function(OrgDb, keytype) {
    kk <- keys(OrgDb, keytype = keytype)
    message("preparing gene to GO mapping data...")
    goAnno <- suppressMessages(
        select(OrgDb,
        keys = kk, keytype = keytype,
        columns = c("GO", "ONTOLOGY")
        )
    )

    goAnno <- goAnno[!is.na(goAnno$GO), ]
    return(goAnno)
}

#' Class "GOSemSimDATA"
#' This class stores IC and gene to go mapping for semantic similarity measurement
#'
#'
#' @name GOSemSimDATA-class
#' @aliases GOSemSimDATA-class show,GOSemSimDATA-method
#'
#' @docType class
#' @slot keys gene ID
#' @slot ont ontology
#' @slot IC IC data
#' @slot geneAnno gene to GO mapping
#' @slot tcssdata tcssdata
#' @slot metadata metadata
#' @exportClass GOSemSimDATA
#' @keywords classes
#' @importFrom methods setClass
setClass("GOSemSimDATA",
  representation = representation(
    keys = "character",
    ont = "character",
    IC = "numeric",
    geneAnno = "data.frame",
    tcssdata = "list",
    metadata = "data.frame"
  )
)

#' @importFrom methods setMethod
setMethod(
  "show", signature(object = "GOSemSimDATA"),
  function(object) {
    cat("#\n# DATA for Semantic Similarity calculation ...\n#\n")
  }
)

library(GOSemSim)

context("Wang")

test_that("Wang's method", {
    hsGO <- godata(annoDb = "org.Hs.eg.db", ont = "MF", computeIC = FALSE)
    x <- goSim("GO:0004022", "GO:0005515", semData=hsGO, measure="Wang")
    expect_true(x >= 0 && x<=1)
})

test_that("Wang semantic value cache is ontology-specific", {
    yulab.utils::initial_cache()
    rel_df <- data.frame(
        go_id = c("GO:test", "GO:test"),
        relationship = c("is_a", "is_a"),
        parent = c("GO:bp_parent", "GO:mf_parent"),
        Ontology = c("BP", "MF"),
        stringsAsFactors = FALSE
    )

    bp_sv <- GOSemSim:::getSV("GO:test", "BP", rel_df)
    mf_sv <- GOSemSim:::getSV("GO:test", "MF", rel_df)

    expect_equal(unname(bp_sv["GO:bp_parent"]), 0.8)
    expect_true(is.na(bp_sv["GO:mf_parent"]))
    expect_equal(unname(mf_sv["GO:mf_parent"]), 0.8)
    expect_true(is.na(mf_sv["GO:bp_parent"]))
})

#' prepare tcss data for TCSS to calculate semantic similarity
#'
#' @param ont ontology
#' @param IC information content
#' @param cutoff the topology cutoff, users can use tcss_cutoff() function to calculate cutoff value  
#'
#' @return list, belonged clusters and its elements for all nodes
#' @noRd
process_tcss <- function(ont, IC, cutoff = NULL) {
    if (length(IC) == 0 || !any(is.finite(IC))) {
        stop("IC data not found, please re-generate your `semData` with `computeIC = TRUE`...")
    }

    if (is.null(cutoff)) {
        message("As cutoff value is not provided, default value based on human will be used")
        cutoff <- switch(ont,
                         MF = 3.5,
                         BP = 3.5,
                         CC = 3.2,
                         DO = 3.5,
                         MPO = 3.5,
                         HPO = 3.5
                         )
    } else if (cutoff <= 0) {
        stop("cutoff value must be positive")
    }

    GO <- names(IC[!is.infinite(IC)])

    offspring <- getOffsprings(ont) 
    # calculate ICT
    ICT <- computeICT(GO, offspring = offspring)
    # nodes smaller than cutoff are meta-terms
    meta_terms <- create_meta_terms(ICT = ICT, cutoff = cutoff)
    # if two parent-child nodes' ICT value too close
    meta_terms <- remove_close(meta_terms, ont = ont, ICT = ICT)
    # relationship between cluster-id and its elements
    meta_graph <- create_sub_terms(meta_terms, offspring = offspring)

    # get the max IC value for each graph
    meta_maxIC <- calc_maxIC(meta_graph, IC = IC)
    ica <- lapply(seq_along(meta_graph), function(i) {
        IC[meta_graph[[i]]] / meta_maxIC[i]
    })
    names(ica) <- meta_terms

    aa <- utils::stack(meta_graph)
    bb <- split(as.character(aa$ind), aa$values)
    clusid <- bb[GO]

    res <- list(
        ## meta_terms == names(meta_graph)
        meta_graph = meta_graph,
        ica = ica,
        clusid = clusid
    )

    return(res)
}

#' compute ICT (Topology Information Content) for each term
#'
#' @param GO character, all go terms, species specific
#' @param offspring list, offspring nodes
#'
#' @return numeric, ICT value
#' @noRd
#'
computeICT <- function(GO, offspring) {
    filtered_offspring <- offspring[GO]
    all <- length(GO)
    num <- -log10(1 / all)

    res <- vapply(filtered_offspring, function(off) {
        if (any(is.na(off))) {
            ## only term itself
            num
        } else {
            ## add term itself
            -log10((sum(off %in% GO) + 1) / all)
        }
    }, numeric(1))
    names(res) <- GO
    return(res)
}


#' all nodes with ICT value under cutoff are meta_terms
#'
#' @param ICT numeric, ICT value with corresponding GO term as name attributes
#' @param cutoff numeric, topological cutoff
#'
#' @return character, sub-graph-root nodes
#' @noRd
create_meta_terms <- function(ICT, cutoff) {
    res <- ICT[ICT <= cutoff]
    names(res)
}

#' calculate every graph's max IC value
#'
#' @param meta_terms character, all cluster ids but "meta"
#' @param GO_element the contained elements
#' @param IC numeric, ICT value
#'
#' @return numeric, max IC in different graphs
#' @noRd
#'
calc_maxIC <- function(meta_graph, IC) {
    # mic : max IC value of all terms
    mic <- max(IC[is.finite(IC)])

    meta_maxIC <- vapply(meta_graph,
                         function(t) {
                                        #all the IC value of elements
                             all <- IC[t]
                                        #all <- all[!is.infinite(all)]
                                        # all <- all[!is.infinite(all) & !is.na(all)]
                             all <- all[is.finite(all)]
                                        # if value is empty, assign the mic value
                             if (length(all) == 0) mic else max(all)
                         }, numeric(1))
    return(meta_maxIC)
}

#' get the relationship between clusters and elements
#'
#' @param meta_terms character, all cluster ids but "meta"
#' @param offspring list
#'
#' @return data.frame, relationship between clusters and elements
#' @noRd
#'
create_sub_terms <- function(meta_terms, offspring) {
    #all terms within cluster
    element <- lapply(meta_terms, function(term) {
        #when no offspring, all shows NA
        all <- offspring[[term]]
        # other sub-root-node
        other_sub <- intersect(all, meta_terms)
        # other sub-root-node's offspring
        other_sub_offs <- unlist(offspring[other_sub])
        # remove
        terms <- setdiff(all, c(other_sub_offs, other_sub))
        # add term itself
        terms <- c(terms, term)
    })
    ## #add "meta" cluster
    ## element$"meta" <- meta_terms
    ## #to unfold into a data.frame
    ## len <- lapply(element, length)

    ## res <- data.frame(cluster = rep(c(meta_terms, "meta"), times = len),
    ##                   element = unlist(element))
    ## return(na.omit(res))

    names(element) <- meta_terms
    return(element)
}

#' remove close relation in meta_terms
#'
#' @param meta_terms character, all cluster ids but "meta"
#' @param ont ontology
#' @param ICT numeric, topological information content value
#'
#' @return character, meta_terms with fewer nodes
#' @noRd
#'
remove_close <- function(meta_terms, ont, ICT) {
    parents <- getParents(ont) 

    # reserve all nodes in advance
    all_ <- meta_terms
    for (term1 in meta_terms) {
        # parent term
        obj <- intersect(parents[[term1]], all_)
        for (term2 in obj) {
            if (ICT[term2] != 0 && ICT[term1] / ICT[term2] <= 1.2) {
                # remove when satisfying the condition
                meta_terms <- setdiff(meta_terms, term1)
                break
            }
        }
    }
    # return the left nodes
    return(meta_terms)
}

#' @keywords internal
"_PACKAGE"



##'Information content of GO terms
##'
##'These datasets are the information contents of GOterms.
##'
##'
##'@name go_term_table
##'@aliases GO gotbl
##'@docType data
##'@references Yu et al. (2010) GOSemSim: an R package for measuring semantic
##'similarity among GO terms and gene products \emph{Bioinformatics} (Oxford,
##'England), 26:7 976--978, April 2010. ISSN 1367-4803
##'\url{http://bioinformatics.oxfordjournals.org/cgi/content/abstract/26/7/976}
##'PMID: 20179076
##'@keywords datasets
NULL

#' Given two GO IDs, calculate their semantic similarity.
#'
#'
#' @title Semantic similarity between two GO terms
#' @param GOID1 GO ID 1
#' @param GOID2 GO ID 2
#' @template params-measure
#' @return similarity
#' @seealso [goSim()] [mgoSim()] [geneSim()] [mgeneSim()] [clusterSim()] [mclusterSim()]
#' @export
#' @examples
#' d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
#' goSim("GO:0004022", "GO:0005515", semData = d, measure = "Wang")
#' @author Guangchuang Yu <https://yulab-smu.top>
goSim <- function(GOID1, GOID2, semData, measure="Wang") {
    res <- termSim(GOID1, GOID2, semData, method=measure)
    res <- as.numeric(res)
    return(round(res, digits = 3))
}

wangMethod <- function(t1, t2, ont) {
    n1 <- length(t1)
    n2 <- length(t2)
    if (n2 > 256) {
        res <- matrix(NA_real_, nrow = n1, ncol = n2, dimnames = list(t1, t2))
        step <- 256L
        for (start in seq.int(1L, n2, by = step)) {
            end <- min(start + step - 1L, n2)
            block <- t2[start:end]
            cols <- lapply(block, function(b) vapply(t1, function(a) wangMethod_internal(a, b, ont = ont), numeric(1)))
            res[, start:end] <- do.call(cbind, cols)
        }
        return(res)
    }
    matrix(mapply(wangMethod_internal,
                  rep(t1, n2),
                  rep(t2, each = n1),
                  MoreArgs = list(ont = ont)),
           dimnames = list(t1, t2), ncol = n2)
}


#' Method Wang for semantic similarity measuring
#'
#' @title wangMethod
#' @param ID1 Ontology Term
#' @param ID2 Ontology Term
#' @param ont Ontology
#' @return semantic similarity score
#' @export
#' @author Guangchuang Yu <https://yulab-smu.top>
wangMethod_internal <- function(ID1, ID2, ont="BP") {
    if (ID1 == ID2)
        return (sim=1)

    if (is_supported_go(ont)) {
        rel_df <- ensure_gotbl_cached()
    } else if (is_supported_do(ont)) {
        rel_df <- get_rel_df(ont)
    } else {
        .meshesEnv <- get(".meshesEnv", envir=.GlobalEnv)
        rel_df <- get("meshtbl", envir=.meshesEnv)
    } 
    
    
    sv.a <- getSV(ID1, ont, rel_df)
    sv.b <- getSV(ID2, ont, rel_df)

    if(all(is.na(sv.a)) || all(is.na(sv.b)))
        return (NA)

    idx         <- intersect(names(sv.a), names(sv.b))
    inter.sva   <- sv.a[idx]
    inter.svb   <- sv.b[idx]
    if (is.null(inter.sva) ||
        is.null(inter.svb) ||
        length(inter.sva) == 0 ||
        length(inter.svb) ==0) {
        return (NA)
    } 
    
    sim <- sum(inter.sva,inter.svb) / sum(sv.a, sv.b)
    return(sim)
}

get_rel_df <- function(ont) {
    ontbl <- sprintf("%stbl", ont)
    get_gosemsim_env()

    res <- yulab.utils::get_cache_element(".GOSemSimEnv", ontbl)
    if (!is.null(res)) return(res)

    ont_db <- load_onto(ont)
    gtb <- toTable(ont_db)
    gtb <- gtb[,1, drop=FALSE]
    gtb <- unique(gtb)

    id <- gtb$id
    parent <- getParents(ont)
    pid <- parent[id]
    cid <- rep(names(pid), times=sapply(pid, length))

    ptb <- data.frame(id=cid,
                      relationship = 'other',
                      parent = unlist(pid),
                      Ontology = ont,
                      stringsAsFactors = FALSE)

    rel_df <- merge(gtb, ptb, by="id")
    rel_df <- rel_df[!is.na(rel_df$id), ]
    rel_df <- rel_df[!is.na(rel_df$parent), ]

    e <- list()
    e[[ontbl]] <- rel_df
    yulab.utils::update_cache_item(".GOSemSimEnv", e)
    return(rel_df)
}


getSV <- function(ID, ont, rel_df, weight=NULL) {
    if (ont == "HDO") {
        topNode <- "DOID:4"
    } else if (ont == "MPO") {
       topNode <- "MP:0000001"
    } else {
        topNode <- "all"
    }
    
    if (ID == topNode) {
        sv <- 1
        names(sv) <- topNode
        return (sv)
    }
    
    if (is.null(weight)) {
        weight <- c(0.8, 0.6, 0.7)
        names(weight) <- c("is_a", "part_of", "other")
    }

    weight_key <- paste(names(weight), weight, sep = "=", collapse = ";")
    cache_key <- paste(ID, ont, weight_key, sep = "|")
    sv <- yulab.utils::get_cache_element("GOSemSim_SemSimCache", cache_key)
    if (!is.null(sv)) return(sv)

    rel_df <- rel_df[rel_df$Ontology == ont,]
    if (! 'relationship' %in% colnames(rel_df))
        rel_df$relationship <- "other"

    ## GO.db emits relationship names as "isa" / "part of" (see GOBPPARENTS),
    ## but the weight table above is keyed on "is_a" / "part_of". Without
    ## normalising these spellings, every edge fails the match below and is
    ## remapped to "other", so Wang similarity silently uses a uniform 0.7
    ## weight for all edges instead of is_a = 0.8 / part_of = 0.6.
    rel_df$relationship[rel_df$relationship %in% c("isa", "is a")] <- "is_a"
    rel_df$relationship[rel_df$relationship %in% c("part of")]     <- "part_of"
    rel_df$relationship[!rel_df$relationship %in% c("is_a", "part_of")] <- "other"


    sv <- 1
    names(sv) <- ID
    allid <- ID

    idx <- which(rel_df[,1] %in% ID)
    while (length(idx) != 0) {
        p <- rel_df[idx,]
        pid <- p$parent
        allid <- c(allid, pid)
        
        sv <- c(sv, weight[p$relationship]*sv[p[,1]])
        names(sv) <- allid
        idx <- which(rel_df[,1] %in% pid)
    }

    sv <- sv[!is.na(names(sv))]
    sv <- sv[!duplicated(names(sv))]

    if (!(ont %in% c("DO", "MPO")))
        sv[topNode] <- 0

    e <- list()
    e[[cache_key]] <- sv
    yulab.utils::update_cache_item("GOSemSim_SemSimCache", e)
    
    return(sv)
}

---
title: "GO Semantic Similarity Analysis"
format: 
    html:
        theme: none
        embed-resources: true
        fontsize: 1.1em
        linestretch: 1.7
author: "Guangchuang Yu\

        School of Basic Medical Sciences, Southern Medical University"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{GOSemSim}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r style, echo=FALSE, results="asis", message=FALSE}
knitr::opts_chunk$set(tidy = FALSE,
                      warning = FALSE,
                      message = FALSE)
```

```{r echo=FALSE, results="hide", message=FALSE}
Biocpkg <- yulab.utils::Biocpkg
```

# Vignette

`r Biocpkg('GOSemSim')` provides methods and utilities for semantic similarity analysis of Gene Ontology annotations, including:

+ semantic similarity measurement among GO terms, gene products, and gene clusters;
+ support for multiple similarity methods, including Wang, Resnik, Lin, Rel, Jiang, and TCSS;
+ preparation of GO annotation data from OrgDb objects or user-supplied annotation data frames;
+ score combination strategies for comparing sets of GO terms or genes.

Please go to <https://yulab-smu.top/biomedical-knowledge-mining-book/> for the full vignette.


# Citation


If you use `r Biocpkg('GOSemSim')` in published research, please cite the most appropriate paper(s) from this list:

1. **Yu G**. [Gene Ontology Semantic Similarity Analysis Using GOSemSim](http://dx.doi.org/10.1007/978-1-0716-0301-7_11). In: Kidder B. (eds) Stem Cell Transcriptional Networks. **_Methods in Molecular Biology_**, 2020, 2117:207-215. Humana, New York, NY. 
2. **Yu G**^#^, Li F^#^, Qin Y, Bo X^\*^, Wu Y and Wang S^\*^. [GOSemSim: an R package for measuring semantic similarity among GO terms and gene products](http://dx.doi.org/10.1093/bioinformatics/btq064). **_Bioinformatics_**. 2010, 26(7):976-978. 



# Need helps?


For questions, please post to [Bioconductor support site](https://support.bioconductor.org/) and tag your post with *GOSemSim*.

---
title: "GO Semantic Similarity Analysis"
author: 
- name: Guangchuang Yu
  email: guangchuangyu@gmail.com
  affiliation: Department of Bioinformatics, School of Basic Medical Sciences, Southern Medical University
date: "`r Sys.Date()`"
output:
  prettydoc::html_pretty:
    toc: true
    theme: cayman
    highlight: github
  pdf_document:
    toc: true
vignette: >
  % \VignetteIndexEntry{GOSemSim}
  % \VignetteEngine{knitr::rmarkdown}
  % \usepackage[utf8]{inputenc}
  %\VignetteEncoding{UTF-8}
---

```{r style, echo=FALSE, results="asis", message=FALSE}
knitr::opts_chunk$set(tidy = FALSE,
                      warning = FALSE,
                      message = FALSE)
```

```{r echo=FALSE, results="hide", message=FALSE}
Biocpkg <- function (pkg) {
    sprintf("[%s](http://bioconductor.org/packages/%s)", pkg, pkg)
}
```

# Vignette

Please go to <https://yulab-smu.top/biomedical-knowledge-mining-book/> for the full vignette.


# Citation


If you use `r Biocpkg('GOSemSim')` in published research, please cite the most appropriate paper(s) from this list:

1. **Yu G**. [Gene Ontology Semantic Similarity Analysis Using GOSemSim](http://dx.doi.org/10.1007/978-1-0716-0301-7_11). In: Kidder B. (eds) Stem Cell Transcriptional Networks. **_Methods in Molecular Biology_**, 2020, 2117:207-215. Humana, New York, NY. 
2. **Yu G**^#^, Li F^#^, Qin Y, Bo X^\*^, Wu Y and Wang S^\*^. [GOSemSim: an R package for measuring semantic similarity among GO terms and gene products](http://dx.doi.org/10.1093/bioinformatics/btq064). **_Bioinformatics_**. 2010, 26(7):976-978. 



# Need helps?


For questions, please post to [Bioconductor support site](https://support.bioconductor.org/) and tag your post with *GOSemSim*.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TCSScutoff.R
\name{tcss_cutoff}
\alias{tcss_cutoff}
\title{determine the topological cutoff for TCSS method}
\usage{
tcss_cutoff(
  OrgDb = NULL,
  keytype = "ENTREZID",
  ont,
  combine_method = "max",
  ppidata
)
}
\arguments{
\item{OrgDb}{OrgDb object}

\item{keytype}{keytype}

\item{ont}{ontology : "BP", "MF", "CC"}

\item{combine_method}{one of "max", "BMA", "avg", "rcmax", "rcmax.avg"}

\item{ppidata}{A data.frame contains positive set and negative set.
Positive set is PPI pairs that already verified.
ppidata has three columns, column 1 and 2 are character, column 3
must be logical value:TRUE/FALSE.}
}
\value{
numeric, topological cutoff for TCSS subgraph construction. The
returned value can be passed to \code{\link[=godata]{godata()}} via the \code{cutoff} argument together
with \code{processTCSS = TRUE}.
}
\description{
determine the topological cutoff for TCSS method
}
\examples{
\dontrun{
    library(org.Hs.eg.db)
    library(STRINGdb)

    string_db <- STRINGdb$new(version = "11.0", species = 9606,
    score_threshold = 700)
    string_proteins <- string_db$get_proteins()

    #get relationship
    ppi <- string_db$get_interactions(string_proteins$protein_external_id)

    ppi$from <- vapply(ppi$from, function(e)
                       strsplit(e, "9606.")[[1]][2], character(1))
    ppi$to <- vapply(ppi$to, function(e)
                       strsplit(e, "9606.")[[1]][2], character(1))
    len <- nrow(ppi)

    #select length
    s_len <- 100
    pos_1 <- sample(len, s_len, replace = T)
    #negative set
    pos_2 <- sample(len, s_len, replace = T)
    pos_3 <- sample(len, s_len, replace = T)
    #union as ppidata
    ppidata <- data.frame(pro1 = c(ppi$from[pos_1], ppi$from[pos_2]),
     pro2 = c(ppi$to[pos_1], ppi$to[pos_3]),
     label = c(rep(TRUE, s_len), rep(FALSE, s_len)),
     stringsAsFactors = FALSE)

    cutoff <- tcss_cutoff(OrgDb = org.Hs.eg.db, keytype = "ENSEMBLPROT",
    ont = "BP", combine_method = "max", ppidata)

    semData <- godata(annoDb = org.Hs.eg.db, keytype = "ENSEMBLPROT",
    ont = "BP", computeIC = TRUE, processTCSS = TRUE, cutoff = cutoff)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/termSim.R
\name{termSim}
\alias{termSim}
\title{termSim}
\usage{
termSim(
  t1,
  t2,
  semData,
  method = c("Wang", "Resnik", "Rel", "Jiang", "Lin", "TCSS")
)
}
\arguments{
\item{t1}{Term vector}

\item{t2}{Term vector}

\item{semData}{GOSemSimDATA object}

\item{method}{One of "Wang", "Resnik", "Rel", "Jiang", "Lin", "TCSS"}
}
\value{
score matrix
}
\description{
Measure similarities between two term vectors.
}
\details{
Provide two term vectors, this function calculates their similarities.

The TCSS method requires \code{semData} generated by \code{\link[=godata]{godata()}} with
\code{computeIC = TRUE} and \code{processTCSS = TRUE}.
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/godata.R
\docType{class}
\name{GOSemSimDATA-class}
\alias{GOSemSimDATA-class}
\alias{show,GOSemSimDATA-method}
\title{Class "GOSemSimDATA"
This class stores IC and gene to go mapping for semantic similarity measurement}
\description{
Class "GOSemSimDATA"
This class stores IC and gene to go mapping for semantic similarity measurement
}
\section{Slots}{

\describe{
\item{\code{keys}}{gene ID}

\item{\code{ont}}{ontology}

\item{\code{IC}}{IC data}

\item{\code{geneAnno}}{gene to GO mapping}

\item{\code{tcssdata}}{tcssdata}

\item{\code{metadata}}{metadata}
}}

\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mclusterSim.R
\name{mclusterSim}
\alias{mclusterSim}
\title{Pairwise semantic similarities for a list of gene clusters}
\usage{
mclusterSim(clusters, semData, measure = "Wang", drop = "IEA", combine = "BMA")
}
\arguments{
\item{clusters}{A list of gene clusters}

\item{semData}{GOSemSimDATA object}

\item{measure}{One of "Resnik", "Lin", "Rel", "Jiang", "TCSS" and "Wang" methods.}

\item{drop}{Evidence codes to drop; use \code{NULL} to keep all GO annotations}

\item{combine}{One of "max", "avg", "rcmax", "BMA" methods, used to combine multiple term scores.}
}
\value{
similarity matrix
}
\description{
Calculate pairwise semantic similarities for a list of gene clusters.
}
\examples{
d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
cluster1 <- c("835", "5261", "241")
cluster2 <- c("578", "582")
cluster3 <- c("307", "308", "317")
clusters <- list(a = cluster1, b = cluster2, c = cluster3)
mclusterSim(clusters, semData = d, measure = "Wang")
}
\seealso{
\code{\link[=goSim]{goSim()}} \code{\link[=mgoSim]{mgoSim()}} \code{\link[=geneSim]{geneSim()}} \code{\link[=mgeneSim]{mgeneSim()}} \code{\link[=clusterSim]{clusterSim()}} \code{\link[=mclusterSim]{mclusterSim()}}
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/godata.R
\name{godata}
\alias{godata}
\title{godata}
\usage{
godata(
  OrgDb = NULL,
  annoDb = NULL,
  keytype = "ENTREZID",
  ont,
  computeIC = TRUE,
  processTCSS = FALSE,
  cutoff = NULL
)
}
\arguments{
\item{OrgDb}{OrgDb object (will be removed in future, please use annoDb instead)}

\item{annoDb}{GO annotation database,
can be OrgDb or a data.frame contains three columns of 'GENE', 'GO' and 'ONTOLOGY'.}

\item{keytype}{keytype}

\item{ont}{one of 'BP', 'MF', 'CC'}

\item{computeIC}{logical, whether computer IC}

\item{processTCSS}{logical, whether to prepare TCSS data. TCSS requires
\code{computeIC = TRUE}, which is enabled automatically when \code{processTCSS = TRUE}.}

\item{cutoff}{topology cutoff for TCSS subgraph construction. If \code{NULL},
empirical defaults are used; users can estimate a cutoff with \code{\link[=tcss_cutoff]{tcss_cutoff()}}.}
}
\value{
GOSemSimDATA object
}
\description{
prepare GO DATA for measuring semantic similarity
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geneSim.R
\name{geneSim}
\alias{geneSim}
\title{Semantic similarity between two genes}
\usage{
geneSim(gene1, gene2, semData, measure = "Wang", drop = "IEA", combine = "BMA")
}
\arguments{
\item{gene1}{Entrez gene ID}

\item{gene2}{Another Entrez gene ID}

\item{semData}{GOSemSimDATA object}

\item{measure}{One of "Resnik", "Lin", "Rel", "Jiang", "TCSS" and "Wang" methods.}

\item{drop}{Evidence codes to drop; use \code{NULL} to keep all GO annotations}

\item{combine}{One of "max", "avg", "rcmax", "BMA" methods, used to combine multiple term scores.}
}
\value{
A list containing similarity value and corresponding GO terms
}
\description{
Given two genes, calculate their semantic similarity and return the score with corresponding GO terms.
}
\examples{
d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
geneSim("241", "251", semData = d, measure = "Wang")
}
\seealso{
\code{\link[=goSim]{goSim()}} \code{\link[=mgoSim]{mgoSim()}} \code{\link[=mgeneSim]{mgeneSim()}} \code{\link[=clusterSim]{clusterSim()}} \code{\link[=mclusterSim]{mclusterSim()}}
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combineMethods.R
\name{combineScores}
\alias{combineScores}
\title{combining similarity matrix to similarity score}
\usage{
combineScores(SimScores, combine)
}
\arguments{
\item{SimScores}{similarity matrix}

\item{combine}{combine method}
}
\value{
similarity value
}
\description{
Functions for combining similarity matrix to similarity score
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GOSemSim-package.R
\docType{package}
\name{GOSemSim-package}
\alias{GOSemSim}
\alias{GOSemSim-package}
\title{GOSemSim: GO-terms Semantic Similarity Measures}
\description{
The semantic comparisons of Gene Ontology (GO) annotations provide quantitative ways to compute similarities between genes and gene groups, and have became important basis for many bioinformatics analysis approaches. GOSemSim is an R package for semantic similarity computation among GO terms, sets of GO terms, gene products and gene clusters. GOSemSim implemented five methods proposed by Resnik, Schlicker, Jiang, Lin and Wang respectively.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://yulab-smu.top/biomedical-knowledge-mining-book/}
  \item Report bugs at \url{https://github.com/YuLab-SMU/GOSemSim/issues}
}

}
\author{
\strong{Maintainer}: Guangchuang Yu \email{guangchuangyu@gmail.com}

Authors:
\itemize{
  \item Guangchuang Yu \email{guangchuangyu@gmail.com}
}

Other contributors:
\itemize{
  \item Alexey Stukalov \email{astukalov@gmail.com} [contributor]
  \item Pingfan Guo \email{1178431277@qq.com} [contributor]
  \item Chuanle Xiao \email{xiaochuanle@126.com} [contributor]
  \item Lluís Revilla Sancho \email{lluis.revilla@gmail.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parseGAF.R
\name{read.gaf}
\alias{read.gaf}
\alias{parse_gff}
\title{read.gaf}
\usage{
read.gaf(file, asis = FALSE, add_indirect_GO = FALSE)

parse_gff(file, asis = FALSE, add_indirect_GO = FALSE)
}
\arguments{
\item{file}{GAF file}

\item{asis}{logical, whether output the original contains of the file and only works if 'add_indirect_GO = FALSE'}

\item{add_indirect_GO}{whether to add indirect GO annotation}
}
\value{
A data.frame. Original table if 'asis' works, otherwise contains 3 conlumns of 'GENE', 'GO' and 'ONTOLOGY'
}
\description{
parse GAF files
}
\details{
given a GAF file, this function extracts the information from it
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildGOmap.R
\name{buildGOmap}
\alias{buildGOmap}
\title{buildGOmap}
\usage{
buildGOmap(TERM2GENE)
}
\arguments{
\item{TERM2GENE}{data.frame with two or three columns of GO TERM, GENE and ONTOLOGY (optional)}
}
\value{
data.frame, GO annotation with direct and indirect annotation
}
\description{
Addding indirect GO annotation
}
\details{
provided by a data.frame of GO TERM (column 1), GENE (column 2) and ONTOLOGY (optional) that
describes GO direct annotation,
this function will add indirect GO annotation of genes.
}
\author{
Yu Guangchuang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{get_organism}
\alias{get_organism}
\title{Get organism name from OrgDb object}
\usage{
get_organism(object)
}
\arguments{
\item{object}{OrgDb object or OrgDb package name}
}
\value{
Organism name
}
\description{
Get organism name from OrgDb object
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OntDb.R
\name{load_onto}
\alias{load_onto}
\title{Load Ontology Database}
\usage{
load_onto(onto = "HDO")
}
\arguments{
\item{onto}{character. The ontology to load (e.g., "HDO").}
}
\value{
An \code{AnnotationDb} object.
}
\description{
Load Ontology Database
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WangMethod.R
\name{wangMethod_internal}
\alias{wangMethod_internal}
\title{wangMethod}
\usage{
wangMethod_internal(ID1, ID2, ont = "BP")
}
\arguments{
\item{ID1}{Ontology Term}

\item{ID2}{Ontology Term}

\item{ont}{Ontology}
}
\value{
semantic similarity score
}
\description{
Method Wang for semantic similarity measuring
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/goSim.R
\name{goSim}
\alias{goSim}
\title{Semantic similarity between two GO terms}
\usage{
goSim(GOID1, GOID2, semData, measure = "Wang")
}
\arguments{
\item{GOID1}{GO ID 1}

\item{GOID2}{GO ID 2}

\item{semData}{GOSemSimDATA object}

\item{measure}{One of "Resnik", "Lin", "Rel", "Jiang", "TCSS" and "Wang" methods.}
}
\value{
similarity
}
\description{
Given two GO IDs, calculate their semantic similarity.
}
\examples{
d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
goSim("GO:0004022", "GO:0005515", semData = d, measure = "Wang")
}
\seealso{
\code{\link[=goSim]{goSim()}} \code{\link[=mgoSim]{mgoSim()}} \code{\link[=geneSim]{geneSim()}} \code{\link[=mgeneSim]{mgeneSim()}} \code{\link[=clusterSim]{clusterSim()}} \code{\link[=mclusterSim]{mclusterSim()}}
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GOSemSim-package.R
\docType{data}
\name{go_term_table}
\alias{go_term_table}
\alias{GO}
\alias{gotbl}
\title{Information content of GO terms}
\description{
These datasets are the information contents of GOterms.
}
\references{
Yu et al. (2010) GOSemSim: an R package for measuring semantic
similarity among GO terms and gene products \emph{Bioinformatics} (Oxford,
England), 26:7 976--978, April 2010. ISSN 1367-4803
\url{http://bioinformatics.oxfordjournals.org/cgi/content/abstract/26/7/976}
PMID: 20179076
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mgoSim.R
\name{mgoSim}
\alias{mgoSim}
\title{Semantic similarity between two GO term sets}
\usage{
mgoSim(GO1, GO2, semData, measure = "Wang", combine = "BMA")
}
\arguments{
\item{GO1}{A set of GO terms}

\item{GO2}{Another set of GO terms}

\item{semData}{GOSemSimDATA object}

\item{measure}{One of "Resnik", "Lin", "Rel", "Jiang", "TCSS" and "Wang" methods.}

\item{combine}{One of "max", "avg", "rcmax", "BMA" methods, used to combine multiple term scores.}
}
\value{
similarity
}
\description{
Given two sets of GO terms, calculate their semantic similarity.
}
\examples{
d <- godata('org.Hs.eg.db', ont = "MF", computeIC = FALSE)
go1 <- c("GO:0004022", "GO:0004024", "GO:0004023")
go2 <- c("GO:0009055", "GO:0020037")
mgoSim("GO:0003824", go2, semData = d, measure = "Wang")
mgoSim(go1, go2, semData = d, measure = "Wang")
}
\seealso{
\code{\link[=goSim]{goSim()}} \code{\link[=mgoSim]{mgoSim()}} \code{\link[=geneSim]{geneSim()}} \code{\link[=mgeneSim]{mgeneSim()}} \code{\link[=clusterSim]{clusterSim()}} \code{\link[=mclusterSim]{mclusterSim()}}
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ICMethods.R
\name{infoContentMethod}
\alias{infoContentMethod}
\title{information content based methods}
\usage{
infoContentMethod(ID1, ID2, method, godata)
}
\arguments{
\item{ID1}{Ontology Term}

\item{ID2}{Ontology Term}

\item{method}{one of "Resnik", "Jiang", "Lin" and "Rel", "TCSS".}

\item{godata}{GOSemSimDATA object}
}
\value{
semantic similarity score
}
\description{
Information Content Based Methods for semantic similarity measuring
}
\details{
implemented for methods proposed by Resnik, Jiang, Lin and Schlicker.
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readBlast2go.R
\name{read.blast2go}
\alias{read.blast2go}
\title{read.blast2go}
\usage{
read.blast2go(file, add_indirect_GO = FALSE)
}
\arguments{
\item{file}{BLAST2GO file}

\item{add_indirect_GO}{whether add indirect GO annotation}
}
\value{
a data frame with three columns: GENE, GO and ONTOLOGY
}
\description{
given a BLAST2GO file, this function extracts the information from it and make it use for TERM2GENE.
}
